package org.jastadd.ragconnect.tests;

import defaultOnlyRead.ast.A;
import defaultOnlyRead.ast.BoxedTypes;
import defaultOnlyRead.ast.MqttHandler;
import defaultOnlyRead.ast.NativeTypes;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.concurrent.TimeUnit;

import static org.jastadd.ragconnect.tests.TestUtils.mqttUri;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

/**
 * Test case "defaultOnlyRead".
 *
 * @author rschoene - Initial contribution
 */
public class DefaultOnlyReadTest extends AbstractMqttTest {

  private static final String TOPIC_NATIVE_INT = "native/int";
  private static final String TOPIC_NATIVE_SHORT = "native/short";
  private static final String TOPIC_NATIVE_LONG = "native/long";
  private static final String TOPIC_NATIVE_FLOAT = "native/float";
  private static final String TOPIC_NATIVE_DOUBLE = "native/double";
  private static final String TOPIC_NATIVE_CHAR = "native/char";
  private static final String TOPIC_NATIVE_STRING = "native/string";

  private static final String TOPIC_BOXED_INTEGER = "boxed/Integer";
  private static final String TOPIC_BOXED_SHORT = "boxed/Short";
  private static final String TOPIC_BOXED_LONG = "boxed/Long";
  private static final String TOPIC_BOXED_FLOAT = "boxed/Float";
  private static final String TOPIC_BOXED_DOUBLE = "boxed/Double";
  private static final String TOPIC_BOXED_CHARACTER = "boxed/Character";

  private A model;
  private NativeTypes integers;
  private NativeTypes floats;
  private NativeTypes chars;
  private BoxedTypes allBoxed;
  private MqttHandler sender;

  @Override
  protected void createModel() {
    model = new A();
    integers = new NativeTypes();
    model.addNativeTypes(integers);
    floats = new NativeTypes();
    model.addNativeTypes(floats);
    chars = new NativeTypes();
    model.addNativeTypes(chars);
    allBoxed = new BoxedTypes();
    model.addBoxedTypes(allBoxed);
  }

  @Override
  protected void setupReceiverAndConnect(boolean writeCurrentValue) throws IOException {
    model.ragconnectSetupMqttWaitUntilReady(2, TimeUnit.SECONDS);

    integers.connectIntValue(mqttUri(TOPIC_NATIVE_INT));
    integers.connectShortValue(mqttUri(TOPIC_NATIVE_SHORT));
    integers.connectLongValue(mqttUri(TOPIC_NATIVE_LONG));
    floats.connectFloatValue(mqttUri(TOPIC_NATIVE_FLOAT));
    floats.connectDoubleValue(mqttUri(TOPIC_NATIVE_DOUBLE));
    chars.connectCharValue(mqttUri(TOPIC_NATIVE_CHAR));
    chars.connectStringValue(mqttUri(TOPIC_NATIVE_STRING));
    allBoxed.connectIntValue(mqttUri(TOPIC_BOXED_INTEGER));
    allBoxed.connectShortValue(mqttUri(TOPIC_BOXED_SHORT));
    allBoxed.connectLongValue(mqttUri(TOPIC_BOXED_LONG));
    allBoxed.connectFloatValue(mqttUri(TOPIC_BOXED_FLOAT));
    allBoxed.connectDoubleValue(mqttUri(TOPIC_BOXED_DOUBLE));
    allBoxed.connectCharValue(mqttUri(TOPIC_BOXED_CHARACTER));

    sender = new MqttHandler().dontSendWelcomeMessage().setHost(TestUtils.getMqttHost());
    assertTrue(sender.waitUntilReady(2, TimeUnit.SECONDS));
  }

  @Override
  protected void communicateSendInitialValue() {
    // empty
  }

  @Override
  protected void communicateOnlyUpdatedValue() throws InterruptedException {
    final int expectedIntValue = 1;
    final short expectedShortValue = 2;
    final long expectedLongValue = 3L;
    final float expectedFloatValue = 4.1f;
    final double expectedDoubleValue = 5.2;
    final char expectedCharValue = 'c';
    final String expectedStringValue = "6.3";

    sender.publish(TOPIC_NATIVE_INT, ByteBuffer.allocate(4).putInt(expectedIntValue).array());
    sender.publish(TOPIC_NATIVE_SHORT, ByteBuffer.allocate(2).putShort(expectedShortValue).array());
    sender.publish(TOPIC_NATIVE_LONG, ByteBuffer.allocate(8).putLong(expectedLongValue).array());
    sender.publish(TOPIC_NATIVE_FLOAT, ByteBuffer.allocate(4).putFloat(expectedFloatValue).array());
    sender.publish(TOPIC_NATIVE_DOUBLE, ByteBuffer.allocate(8).putDouble(expectedDoubleValue).array());
    sender.publish(TOPIC_NATIVE_CHAR, ByteBuffer.allocate(2).putChar(expectedCharValue).array());
    sender.publish(TOPIC_NATIVE_STRING, expectedStringValue.getBytes());

    sender.publish(TOPIC_BOXED_INTEGER, ByteBuffer.allocate(4).putInt(expectedIntValue).array());
    sender.publish(TOPIC_BOXED_SHORT, ByteBuffer.allocate(2).putShort(expectedShortValue).array());
    sender.publish(TOPIC_BOXED_LONG, ByteBuffer.allocate(8).putLong(expectedLongValue).array());
    sender.publish(TOPIC_BOXED_FLOAT, ByteBuffer.allocate(4).putFloat(expectedFloatValue).array());
    sender.publish(TOPIC_BOXED_DOUBLE, ByteBuffer.allocate(8).putDouble(expectedDoubleValue).array());
    sender.publish(TOPIC_BOXED_CHARACTER, ByteBuffer.allocate(2).putChar(expectedCharValue).array());

    TestUtils.waitForMqtt();

    assertEquals(expectedIntValue, integers.getIntValue());
    assertEquals(expectedShortValue, integers.getShortValue());
    assertEquals(expectedLongValue, integers.getLongValue());
    assertEquals(expectedFloatValue, floats.getFloatValue(), TestUtils.DELTA);
    assertEquals(expectedDoubleValue, floats.getDoubleValue(), TestUtils.DELTA);
    assertEquals(expectedCharValue, chars.getCharValue());
    assertEquals(expectedStringValue, chars.getStringValue());

    assertEquals(expectedIntValue, allBoxed.getIntValue().intValue());
    assertEquals(expectedShortValue, allBoxed.getShortValue().shortValue());
    assertEquals(expectedLongValue, allBoxed.getLongValue().longValue());
    assertEquals(expectedFloatValue, allBoxed.getFloatValue(), TestUtils.DELTA);
    assertEquals(expectedDoubleValue, allBoxed.getDoubleValue(), TestUtils.DELTA);
    assertEquals(expectedCharValue, allBoxed.getCharValue().charValue());
  }

  @Override
  public void closeConnections() {
    if (sender != null) {
      sender.close();
    }
    if (model != null) {
      model.ragconnectCloseConnections();
    }
  }

}
