package org.jastadd.ragconnect.tests;

import defaultOnlyWrite.ast.A;
import defaultOnlyWrite.ast.BoxedTypesSyn;
import defaultOnlyWrite.ast.MqttHandler;
import defaultOnlyWrite.ast.NativeTypesSyn;

import java.io.IOException;
import java.util.concurrent.TimeUnit;

import static org.jastadd.ragconnect.tests.TestUtils.mqttUri;
import static org.junit.jupiter.api.Assertions.*;

/**
 * Test case "defaultOnlyRead".
 *
 * @author rschoene - Initial contribution
 */
public class DefaultOnlyWriteTest extends AbstractMqttTest {

  private static final String TOPIC_NATIVE_INT = "native/int";
  private static final String TOPIC_NATIVE_SHORT = "native/short";
  private static final String TOPIC_NATIVE_LONG = "native/long";
  private static final String TOPIC_NATIVE_FLOAT = "native/float";
  private static final String TOPIC_NATIVE_DOUBLE = "native/double";
  private static final String TOPIC_NATIVE_CHAR = "native/char";
  private static final String TOPIC_NATIVE_STRING = "native/string";

  private static final String TOPIC_BOXED_INTEGER = "boxed/Integer";
  private static final String TOPIC_BOXED_SHORT = "boxed/Short";
  private static final String TOPIC_BOXED_LONG = "boxed/Long";
  private static final String TOPIC_BOXED_FLOAT = "boxed/Float";
  private static final String TOPIC_BOXED_DOUBLE = "boxed/Double";
  private static final String TOPIC_BOXED_CHARACTER = "boxed/Character";

  private A model;
  private NativeTypesSyn nativeIntegers;
  private NativeTypesSyn nativeFloats;
  private NativeTypesSyn nativeChars;
  private BoxedTypesSyn boxedIntegers;
  private BoxedTypesSyn boxedFloats;
  private BoxedTypesSyn boxedChars;
  private MqttHandler receiver;
  private ReceiverData data;

  @Override
  protected void createModel() {
    model = new A();

    nativeIntegers = new NativeTypesSyn();
    nativeFloats = new NativeTypesSyn();
    nativeChars = new NativeTypesSyn();
    model.addNativeTypesSyn(nativeIntegers);
    model.addNativeTypesSyn(nativeFloats);
    model.addNativeTypesSyn(nativeChars);

    boxedIntegers = new BoxedTypesSyn();
    boxedFloats = new BoxedTypesSyn();
    boxedChars = new BoxedTypesSyn();
    model.addBoxedTypesSyn(boxedIntegers);
    model.addBoxedTypesSyn(boxedFloats);
    model.addBoxedTypesSyn(boxedChars);

    setData("1", "1.1", "ab");
  }

  @Override
  protected void setupReceiverAndConnect(boolean writeCurrentValue) throws IOException {
    model.ragconnectSetupMqttWaitUntilReady(2, TimeUnit.SECONDS);

    receiver = new MqttHandler().dontSendWelcomeMessage().setHost(TestUtils.getMqttHost());
    assertTrue(receiver.waitUntilReady(2, TimeUnit.SECONDS));

    nativeIntegers.addNativeIntDependency(nativeIntegers);
    nativeIntegers.addNativeShortDependency(nativeIntegers);
    nativeIntegers.addNativeLongDependency(nativeIntegers);
    nativeFloats.addNativeFloatDependency(nativeFloats);
    nativeFloats.addNativeDoubleDependency(nativeFloats);
    nativeChars.addNativeCharDependency(nativeChars);
    nativeChars.addNativeStringDependency(nativeChars);

    boxedIntegers.addBoxedIntDependency(boxedIntegers);
    boxedIntegers.addBoxedShortDependency(boxedIntegers);
    boxedIntegers.addBoxedLongDependency(boxedIntegers);
    boxedFloats.addBoxedFloatDependency(boxedFloats);
    boxedFloats.addBoxedDoubleDependency(boxedFloats);
    boxedChars.addBoxedCharDependency(boxedChars);

    data = new ReceiverData();

    receiver.newConnection(TOPIC_NATIVE_INT, bytes -> {
      data.numberOfNativeIntValues += 1;
      data.lastNativeIntValue = java.nio.ByteBuffer.wrap(bytes).getInt();
    });
    receiver.newConnection(TOPIC_NATIVE_SHORT, bytes -> {
      data.numberOfNativeShortValues += 1;
      data.lastNativeShortValue = java.nio.ByteBuffer.wrap(bytes).getShort();
    });
    receiver.newConnection(TOPIC_NATIVE_LONG, bytes -> {
      data.numberOfNativeLongValues += 1;
      data.lastNativeLongValue = java.nio.ByteBuffer.wrap(bytes).getLong();
    });
    receiver.newConnection(TOPIC_NATIVE_FLOAT, bytes -> {
      data.numberOfNativeFloatValues += 1;
      data.lastNativeFloatValue = java.nio.ByteBuffer.wrap(bytes).getFloat();
    });
    receiver.newConnection(TOPIC_NATIVE_DOUBLE, bytes -> {
      data.numberOfNativeDoubleValues += 1;
      data.lastNativeDoubleValue = java.nio.ByteBuffer.wrap(bytes).getDouble();
    });
    receiver.newConnection(TOPIC_NATIVE_CHAR, bytes -> {
      data.numberOfNativeCharValues += 1;
      data.lastNativeCharValue = java.nio.ByteBuffer.wrap(bytes).getChar();
    });
    receiver.newConnection(TOPIC_NATIVE_STRING, bytes -> {
      data.numberOfNativeStringValues += 1;
      data.lastNativeStringValue = new String(bytes);
    });
    receiver.newConnection(TOPIC_BOXED_INTEGER, bytes -> {
      data.numberOfBoxedIntValues += 1;
      data.lastBoxedIntValue = java.nio.ByteBuffer.wrap(bytes).getInt();
    });
    receiver.newConnection(TOPIC_BOXED_SHORT, bytes -> {
      data.numberOfBoxedShortValues += 1;
      data.lastBoxedShortValue = java.nio.ByteBuffer.wrap(bytes).getShort();
    });
    receiver.newConnection(TOPIC_BOXED_LONG, bytes -> {
      data.numberOfBoxedLongValues += 1;
      data.lastBoxedLongValue = java.nio.ByteBuffer.wrap(bytes).getLong();
    });
    receiver.newConnection(TOPIC_BOXED_FLOAT, bytes -> {
      data.numberOfBoxedFloatValues += 1;
      data.lastBoxedFloatValue = java.nio.ByteBuffer.wrap(bytes).getFloat();
    });
    receiver.newConnection(TOPIC_BOXED_DOUBLE, bytes -> {
      data.numberOfBoxedDoubleValues += 1;
      data.lastBoxedDoubleValue = java.nio.ByteBuffer.wrap(bytes).getDouble();
    });
    receiver.newConnection(TOPIC_BOXED_CHARACTER, bytes -> {
      data.numberOfBoxedCharValues += 1;
      data.lastBoxedCharValue = java.nio.ByteBuffer.wrap(bytes).getChar();
    });

    nativeIntegers.connectIntValue(mqttUri(TOPIC_NATIVE_INT), writeCurrentValue);
    nativeIntegers.connectShortValue(mqttUri(TOPIC_NATIVE_SHORT), writeCurrentValue);
    nativeIntegers.connectLongValue(mqttUri(TOPIC_NATIVE_LONG), writeCurrentValue);
    nativeFloats.connectFloatValue(mqttUri(TOPIC_NATIVE_FLOAT), writeCurrentValue);
    nativeFloats.connectDoubleValue(mqttUri(TOPIC_NATIVE_DOUBLE), writeCurrentValue);
    nativeChars.connectCharValue(mqttUri(TOPIC_NATIVE_CHAR), writeCurrentValue);
    nativeChars.connectStringValue(mqttUri(TOPIC_NATIVE_STRING), writeCurrentValue);
    boxedIntegers.connectIntValue(mqttUri(TOPIC_BOXED_INTEGER), writeCurrentValue);
    boxedIntegers.connectShortValue(mqttUri(TOPIC_BOXED_SHORT), writeCurrentValue);
    boxedIntegers.connectLongValue(mqttUri(TOPIC_BOXED_LONG), writeCurrentValue);
    boxedFloats.connectFloatValue(mqttUri(TOPIC_BOXED_FLOAT), writeCurrentValue);
    boxedFloats.connectDoubleValue(mqttUri(TOPIC_BOXED_DOUBLE), writeCurrentValue);
    boxedChars.connectCharValue(mqttUri(TOPIC_BOXED_CHARACTER), writeCurrentValue);
  }

  @Override
  protected void communicateSendInitialValue() throws InterruptedException {
    // check initial value
    TestUtils.waitForMqtt();
    checkData(1, 1, 1.1, 'a', "ab");

    // set new value
    setData("2", "2.2", "cd");

    // check new value
    TestUtils.waitForMqtt();
    checkData(2, 2, 2.2, 'c', "cd");

    // set new value
    setData("3", "3.2", "ee");

    // check new value
    TestUtils.waitForMqtt();
    checkData(3, 3, 3.2, 'e', "ee");
  }

  @Override
  protected void communicateOnlyUpdatedValue() throws InterruptedException {
    // check initial value (will be default values)
    TestUtils.waitForMqtt();
    checkData(0, null, null, null, null);

    // set new value
    setData("2", "2.2", "cd");

    // check new value
    TestUtils.waitForMqtt();
    checkData(1, 2, 2.2, 'c', "cd");

    // set new value
    setData("3", "3.2", "ee");

    // check new value
    TestUtils.waitForMqtt();
    checkData(2, 3, 3.2, 'e', "ee");
  }

  @Override
  public void closeConnections() {
    if (receiver != null) {
      receiver.close();
    }
    if (model != null) {
      model.ragconnectCloseConnections();
    }
  }

  private void setData(String integerDriver, String floatDriver, String stringDriver) {
    nativeIntegers.setDriverSyn(integerDriver);
    nativeFloats.setDriverSyn(floatDriver);
    nativeChars.setDriverSyn(stringDriver);

    boxedIntegers.setDriverSyn(integerDriver);
    boxedFloats.setDriverSyn(floatDriver);
    boxedChars.setDriverSyn(stringDriver);
  }

  private void checkData(int expectedNumberOfValues,
                         Integer expectedInt, Double expectedDouble,
                         Character expectedChar, String expectedString) {
    assertEquals(expectedNumberOfValues, data.numberOfNativeIntValues);
    assertEquals(expectedNumberOfValues, data.numberOfNativeShortValues);
    assertEquals(expectedNumberOfValues, data.numberOfNativeLongValues);
    assertEquals(expectedNumberOfValues, data.numberOfNativeFloatValues);
    assertEquals(expectedNumberOfValues, data.numberOfNativeDoubleValues);
    assertEquals(expectedNumberOfValues, data.numberOfNativeCharValues);
    assertEquals(expectedNumberOfValues, data.numberOfNativeStringValues);

    assertEquals(expectedNumberOfValues, data.numberOfBoxedIntValues);
    assertEquals(expectedNumberOfValues, data.numberOfBoxedShortValues);
    assertEquals(expectedNumberOfValues, data.numberOfBoxedLongValues);
    assertEquals(expectedNumberOfValues, data.numberOfBoxedFloatValues);
    assertEquals(expectedNumberOfValues, data.numberOfBoxedDoubleValues);
    assertEquals(expectedNumberOfValues, data.numberOfBoxedCharValues);

    if (expectedInt != null) {
      assertEquals(expectedInt.intValue(), data.lastNativeIntValue);
      assertEquals(expectedInt.shortValue(), data.lastNativeShortValue);
      assertEquals(expectedInt.longValue(), data.lastNativeLongValue);
      assertEquals(expectedInt.intValue(), data.lastBoxedIntValue.intValue());
      assertEquals(expectedInt.shortValue(), data.lastBoxedShortValue.shortValue());
      assertEquals(expectedInt.longValue(), data.lastBoxedLongValue.longValue());
    } else {
      assertEquals(0, data.lastNativeIntValue);
      assertEquals(0, data.lastNativeShortValue);
      assertEquals(0, data.lastNativeLongValue);
      assertNull(data.lastBoxedIntValue);
      assertNull(data.lastBoxedShortValue);
      assertNull(data.lastBoxedLongValue);
    }

    if (expectedDouble != null) {
      assertEquals(expectedDouble.floatValue(), data.lastNativeFloatValue, TestUtils.DELTA);
      assertEquals(expectedDouble, data.lastNativeDoubleValue, TestUtils.DELTA);
      assertEquals(expectedDouble.floatValue(), data.lastBoxedFloatValue, TestUtils.DELTA);
      assertEquals(expectedDouble, data.lastBoxedDoubleValue, TestUtils.DELTA);
    } else {
      assertEquals(0f, data.lastNativeFloatValue, TestUtils.DELTA);
      assertEquals(0d, data.lastNativeDoubleValue, TestUtils.DELTA);
      assertNull(data.lastBoxedFloatValue);
      assertNull(data.lastBoxedDoubleValue);
    }

    if (expectedChar != null) {
      assertEquals(expectedChar.charValue(), data.lastNativeCharValue);
      assertEquals(expectedChar, data.lastBoxedCharValue);
    } else {
      assertEquals('\0', data.lastNativeCharValue);
      assertNull(data.lastBoxedCharValue);
    }
    assertEquals(expectedString, data.lastNativeStringValue);
  }

  private static class ReceiverData {
    int lastNativeIntValue;
    int numberOfNativeIntValues = 0;
    short lastNativeShortValue;
    int numberOfNativeShortValues = 0;
    long lastNativeLongValue;
    int numberOfNativeLongValues = 0;
    float lastNativeFloatValue;
    int numberOfNativeFloatValues = 0;
    double lastNativeDoubleValue;
    int numberOfNativeDoubleValues = 0;
    char lastNativeCharValue;
    int numberOfNativeCharValues = 0;
    String lastNativeStringValue;
    int numberOfNativeStringValues = 0;

    Integer lastBoxedIntValue;
    int numberOfBoxedIntValues = 0;
    Short lastBoxedShortValue;
    int numberOfBoxedShortValues = 0;
    Long lastBoxedLongValue;
    int numberOfBoxedLongValues = 0;
    Float lastBoxedFloatValue;
    int numberOfBoxedFloatValues = 0;
    Double lastBoxedDoubleValue;
    int numberOfBoxedDoubleValues = 0;
    Character lastBoxedCharValue;
    int numberOfBoxedCharValues = 0;
  }

}
